﻿using Newtonsoft.Json;
using Newtonsoft.Json.Linq;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Reflection;
using WebSocketSharp;
using WebSocketSharp.Server;


namespace WSCF
{
    public class Reflect_Proxy : WebSocketBehavior,IDisposable
    {
        static NLog.Logger logger = NLog.LogManager.GetCurrentClassLogger();
        //对象的接口类型
        Type interfaceType;
        //代理对象
        object obj;
        //忽略设置
        public bool ignoreSet = false;
        
        List<AnyEvent> anyEvents = new List<AnyEvent>();
        SubPropertyNode rootNode;
        List<string> properties = new List<string>();
        public bool isAlive = false;

        
        public Reflect_Proxy()
        {

        }
        public void Init(Type interfaceType, object obj)
        {
            this.interfaceType = interfaceType;
            this.obj = obj;

            rootNode = new SubPropertyNode
            {
                Obj = obj,
                InterfaceType = interfaceType,
                SubPropertyChanged = Sub_PropertyChanged
            };

            //注册 obj 的PropertyChanged 事件，获取 interfaceType 全部属性名称，包括它的父类的全部属性名称
            InitPropertyChanged();

            //处理[PropertyPush]
            COMMON.InitPropertyPush(rootNode);

            //处理[Push]
            InitEventPush();
        }


        #region Init
        void InitPropertyChanged()
        {
            if (!typeof(INotifyPropertyChanged).IsAssignableFrom(interfaceType))
                return;

            //继承了INotifyPropertyChanged
            ((INotifyPropertyChanged)(this.obj)).PropertyChanged += rootNode.PropertyChanged;

            var propertyInfos = COMMON.GetAllPropertyInfos(this.interfaceType);//获取全部属性， 包括父类的属性

            foreach (var propertyInfo in propertyInfos)
            {

                if (propertyInfo.GetCustomAttribute<JsonIgnoreAttribute>() != null)
                    continue;//这个不需要推送

                if (properties.Contains(propertyInfo.Name))
                    continue;
                properties.Add(propertyInfo.Name);
            }
        }


        void InitEventPush()
        {
            var interfaceTypes = new List<Type>();

            interfaceTypes.Add(this.interfaceType);
            interfaceTypes.AddRange(this.interfaceType.GetInterfaces());


            var eventInfos = new List<EventInfo>();
            foreach (var ifaceType in interfaceTypes)
            {
                eventInfos.AddRange(ifaceType.GetEvents());
            }

            foreach (var eventInfo in eventInfos)
            {
                var pushAttribute = eventInfo.GetCustomAttribute<PushAttribute>();
                if (pushAttribute == null)
                    continue;

                if (anyEvents.Any(ae => ae.eventName == eventInfo.Name))
                    continue;//已经添加了

                var anyEvent = new AnyEvent()
                {
                    eventName = eventInfo.Name,
                    PushObjInfoEx = (rd) => {
                        //数据推送！！！
                        //再嵌套

                        Reflect_OBJ_INTERFACE.PkgData pkgData = new Reflect_OBJ_INTERFACE.PkgData(
                            Reflect_OBJ_INTERFACE.PkgName.PUSH_Event,
                            null,
                            JObject.FromObject(rd)
                        );
                        string json = JsonConvert.SerializeObject(pkgData);
                        SendEx(json);
                    }
                };
                //这个事件需要推送
                eventInfo.AddEventHandler(obj, anyEvent.eventHandler);
                anyEvents.Add(anyEvent);

            }
        }

        void SendEx(string msg) {
            try
            {
                Send(msg);
            }
            catch (Exception e)
            {
                //异常,通常是断开了！！
                logger.Error(e);
            }
        }
        private void Sub_PropertyChanged(SubPropertyNode node, PropertyChangedEventArgs e)
        {
            //if (ignoreSet)//从服务器接收的数据，不用再推送给服务器
            //    return;
            var type = node.Obj.GetType();

            var propertyInfo = type.GetProperty(e.PropertyName);
            if (propertyInfo == null)
                return;//获取失败！！！

            if (propertyInfo.GetCustomAttribute<JsonIgnoreAttribute>() != null)
                return;//这个不需要推送

            var v = propertyInfo.GetValue(node.Obj);
            var jObject = new JObject();
            if (v == null)
                jObject.Add(propertyInfo.Name, null);
            else
                jObject.Add(propertyInfo.Name, JToken.FromObject(v));

            var rData = new Reflect_OBJ_INTERFACE.ReflectData
            {
                name = COMMON.GetNodePath(node),
                data = jObject
            };

            Reflect_OBJ_INTERFACE.PkgData pkgData = new Reflect_OBJ_INTERFACE.PkgData(
                Reflect_OBJ_INTERFACE.PkgName.PUSH_PropertyChanged,
                null,
                JToken.FromObject(rData)
            );
            string json = JsonConvert.SerializeObject(pkgData);
            SendEx(json);
        }
        #endregion


        protected override void OnOpen()
        {
            base.OnOpen();
            isAlive = true;
            AfterOpen?.Invoke(this);            
        }

        public Action<Reflect_Proxy> AfterOpen;

        protected override void OnClose(CloseEventArgs e)
        {
            base.OnClose(e);
            isAlive = false;

            Dispose();
        }

        protected override void OnMessage(MessageEventArgs e)
        {
            if (!e.IsText)
                return;
            string json = e.Data;
            var p = JsonConvert.DeserializeObject<Reflect_OBJ_INTERFACE.PkgData>(json);
            OnMessage(p);
        }

        void OnMessage(Reflect_OBJ_INTERFACE.PkgData pkgData)
        {
            var pkgName = Enum.Parse(typeof(Reflect_OBJ_INTERFACE.PkgName), pkgData.name);
            switch (pkgName)
            {
                case Reflect_OBJ_INTERFACE.PkgName.CALL_GetAllProperties:
                    {

                        var jObject = new JObject();

                        var type = obj.GetType();

                        foreach (var propertyName in properties)
                        {
                            var propertyInfo = type.GetProperty(propertyName);
                            var v = propertyInfo.GetValue(obj);
                            if (v == null)
                                jObject.Add(propertyInfo.Name, null);
                            else
                                jObject.Add(propertyInfo.Name, JToken.FromObject(v));
                        }

                        var rData = new Reflect_OBJ_INTERFACE.ReflectData
                        {
                            data = jObject
                        };


                        pkgData.data = JToken.FromObject(rData);

                        string json = JsonConvert.SerializeObject(pkgData);
                        SendEx(json);
                    }
                    break;
                case Reflect_OBJ_INTERFACE.PkgName.CALL_SetProperty:
                    {
                        //ignoreSet = true;
                        var rData = pkgData.data.ToObject<Reflect_OBJ_INTERFACE.ReflectData>();


                        //sub property
                        var node = COMMON.FindNode(rootNode, rData.name);
                        if (node == null) {
                            //异常
                            //客户端乱发过来
                            break;
                        }

                        string json = rData.data.ToString();
                        JsonConvert.PopulateObject(json, node.Obj);
                        
                        //ignoreSet = false;
                    }
                    break;
                case Reflect_OBJ_INTERFACE.PkgName.CALL_MethodInvoke:
                    {
                        var rData = pkgData.data.ToObject<Reflect_OBJ_INTERFACE.ReflectData>();
                        var type = obj.GetType();
                        var paramNames_req = rData.data.Children().OfType<JProperty>().Select(p => p.Name);
                        MethodInfo methodInfo = GetMethodInfo(type, rData.name, paramNames_req);
                        if (methodInfo == null)
                        {
                            //不能找到，
                            throw new Exception($"程序写错了, 不能找到 rData.name={rData.name} 或者 参数名称不对，没法完全匹配");
                        }

                        var parameterInfos = methodInfo.GetParameters();
                        object[] parameters = new object[parameterInfos.Count()];
                        for (int i = 0; i < parameters.Count(); i++)
                        {
                            var ptype = parameterInfos[i].ParameterType;
                            var pname = parameterInfos[i].Name;
                            if (string.Compare(pname, Reflect_OBJ_INTERFACE.asyncDelegate, true) == 0)
                            {
                                parameters[i] = new AsyncCBHandler(asyncDelegate);
                            }
                            else if (string.Compare(pname, Reflect_OBJ_INTERFACE.asyncContext, true) == 0)
                            {
                                parameters[i] = new CC() { methodName = rData.name, guid = pkgData.guid };
                            }
                            else
                            {
                                parameters[i] = rData.data[pname].ToObject(ptype);
                            }
                        }
                        methodInfo.Invoke(obj, parameters);
                    }break;
             }
            

        }
        MethodInfo GetMethodInfo(Type type, string name, IEnumerable<string> parameterNames)
        {

            var methodInfos = from mi in type.GetMethods() where mi.Name == name select mi;
            if (methodInfos.Count() == 0)
                return null;
            if (methodInfos.Count() == 1)
                return methodInfos.First();

            //必须完全匹配
            foreach (var methodInfo in methodInfos)
            {
                var parameterInfos = methodInfo.GetParameters();
                //全部参数名称
                var names = parameterInfos.Select(pi => pi.Name).ToList();
                //删除掉 asyncDelegate，asyncContext
                names.Remove(Reflect_OBJ_INTERFACE.asyncDelegate);
                names.Remove(Reflect_OBJ_INTERFACE.asyncContext);

                var names_req = parameterNames;

                if (names.Count() != names_req.Count())
                    continue;//数量不一致，肯定不同

                var sames = names_req.Intersect(names);

                if (sames.Count() != names_req.Count())
                    continue;// names 与 names_req 的交集数量与names_req不一样，肯定不同

                //就是它
                return methodInfo;

            }
            return null;
        }
        void asyncDelegate(object asyncContext, object retData)
        {
            var cc = (CC)asyncContext;

            var rData = new Reflect_OBJ_INTERFACE.ReflectData()
            {
                name = cc.methodName,
                data = retData == null ? null : JToken.FromObject(retData)
            };


            Reflect_OBJ_INTERFACE.PkgData pkgData = new Reflect_OBJ_INTERFACE.PkgData(
                Reflect_OBJ_INTERFACE.PkgName.CALL_MethodInvoke,
                cc.guid,
                JToken.FromObject(rData)
            );

            string json = JsonConvert.SerializeObject(pkgData);
            SendEx(json);
        }

        public void Dispose()
        {
            if (typeof(INotifyPropertyChanged).IsAssignableFrom(interfaceType))
            {
                //继承了INotifyPropertyChanged
                ((INotifyPropertyChanged)(this.obj)).PropertyChanged -= rootNode.PropertyChanged;
            }

            //释放subProperties
            COMMON.NodeDispose(rootNode);
        }

        class CC
        {
            public string methodName;
            public string guid;
        }
    
        
    }
    
    /// <summary>
    /// 事件推送包装
    /// </summary>
    class AnyEvent
    {
        /// <summary>
        /// 事件名称
        /// </summary>
        public string eventName;
        /// <summary>
        /// json包装后的ReflectData 推送
        /// </summary>
        public Action<Reflect_OBJ_INTERFACE.ReflectData> PushObjInfoEx;
        /// <summary>
        /// 给 EventInfo 调用
        /// </summary>
        public EventHandler eventHandler;
        public AnyEvent()
        {
            eventHandler = new EventHandler(Obj_AnyEvent);
        }
        void Obj_AnyEvent(object sender, EventArgs e)
        {
            var rData = new Reflect_OBJ_INTERFACE.ReflectData()
            {
                name = eventName,
                data = JObject.FromObject(e)
            };
            PushObjInfoEx?.Invoke(rData);
        }
    }
}
