﻿using Newtonsoft.Json;
using Newtonsoft.Json.Linq;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Net;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;

namespace WSCF
{
    public class Reflect_SeviceClient : WebSocketClient, IDisposable
    {
        /// <summary>
        /// 忽略设置
        /// </summary>
        public bool ignoreSet;
        /// <summary>
        /// 代理对象 接口类型
        /// </summary>
        protected virtual Type InterfaceType { get; }


        class AnyEvent
        {
            public string name;
            public string triggerName;
            public Type retType;
        }
        class AnyCall
        {
            public string name;
            public Type retType;
        }
        class AnyMethodInvoke
        {
            public string guid;
            public AsyncCBHandler asyncDelegate;
            public object asyncContext;
        }

        List<AnyEvent> anyEvents = new List<AnyEvent>();
        List<AnyCall> anyCalls = new List<AnyCall>();
        List<AnyMethodInvoke> anyMethodInvokes = new List<AnyMethodInvoke>();
        SubPropertyNode rootNode;
        List<string> properties = new List<string>();

        public Reflect_SeviceClient(string serverName,bool isByConnName, string addrOrConnName):base(serverName)
        {
            if (isByConnName == false)
            {
                base.Init(addrOrConnName);
            }
            else {
                base.InitByConnName(addrOrConnName);
            }
            init();
        }

        void init()
        {
            rootNode = new SubPropertyNode
            {
                Obj = this,
                InterfaceType = InterfaceType,
                SubPropertyChanged = Sub_PropertyChanged
            };

            InitPropertyChanged();
            COMMON.InitPropertyPush(rootNode);
            InitEventPush();
            InitCall();
        }
        void InitPropertyChanged()
        {
            if (!typeof(INotifyPropertyChanged).IsAssignableFrom(InterfaceType))
                return;

            //继承了INotifyPropertyChanged
            this.PropertyChanged += rootNode.PropertyChanged;

            var propertyInfos = COMMON.GetAllPropertyInfos(this.InterfaceType);//获取全部属性， 包括父类的属性

            foreach (var propertyInfo in propertyInfos)
            {

                if (propertyInfo.GetCustomAttribute<JsonIgnoreAttribute>() != null)
                    continue;//这个不需要推送

                if (properties.Contains(propertyInfo.Name))
                    continue;
                properties.Add(propertyInfo.Name);
            }
        }

        void InitEventPush()
        {
            var interfaceTypes = new List<Type>();

            interfaceTypes.Add(InterfaceType);
            interfaceTypes.AddRange(InterfaceType.GetInterfaces());


            var eventInfos = new List<EventInfo>();
            foreach (var ifaceType in interfaceTypes)
            {
                eventInfos.AddRange(ifaceType.GetEvents());
            }

            foreach (var eventInfo in eventInfos)
            {
                var pushAttribute = eventInfo.GetCustomAttribute<PushAttribute>();
                if (pushAttribute == null)
                    continue;

                if (anyEvents.Any(ae => ae.name == eventInfo.Name))
                    continue;//已经添加了
                string triggerName;
                if (string.IsNullOrEmpty(pushAttribute.TriggerName))
                {
                    triggerName = "Trigger_" + eventInfo.Name;
                }
                else
                {
                    triggerName = pushAttribute.TriggerName;
                }
                var anyEvent = new AnyEvent()
                {
                    name = eventInfo.Name,
                    triggerName = triggerName,
                    retType = pushAttribute.EventArgsType
                };
                anyEvents.Add(anyEvent);

            }
        }

        void InitCall()
        {
            var interfaceTypes = new List<Type>();

            interfaceTypes.Add(InterfaceType);
            interfaceTypes.AddRange(InterfaceType.GetInterfaces());


            var methodInfos = new List<MethodInfo>();
            foreach (var ifaceType in interfaceTypes)
            {
                methodInfos.AddRange(ifaceType.GetMethods());
            }

            foreach (var methodInfo in methodInfos)
            {
                var callAttribute = methodInfo.GetCustomAttribute<CallAttribute>();
                if (callAttribute == null)
                    continue;

                if (anyCalls.Any(ae => ae.name == methodInfo.Name))
                    continue;//已经添加了

                var anyCall = new AnyCall()
                {
                    name = methodInfo.Name,
                    retType = callAttribute.ReponseType
                };
                anyCalls.Add(anyCall);

            }
        }

        private void Sub_PropertyChanged(SubPropertyNode node, PropertyChangedEventArgs e)
        {
            if (!IsConnected)
                return;

            if (ignoreSet)//从服务器接收的数据，不用再推送给服务器
                return;

            var type = node.Obj.GetType();

            var propertyInfo = type.GetProperty(e.PropertyName);
            if (propertyInfo == null)
                return;//获取失败！！！

            if (propertyInfo.GetCustomAttribute<JsonIgnoreAttribute>() != null)
                return;//这个不需要推送

            var v = propertyInfo.GetValue(node.Obj);
            var jObject = new JObject();
            if (v == null)
                jObject.Add(propertyInfo.Name, null);
            else
                jObject.Add(propertyInfo.Name, JToken.FromObject(v));

            var rData = new Reflect_OBJ_INTERFACE.ReflectData
            {
                name = COMMON.GetNodePath(node),
                data = jObject
            };

            Reflect_OBJ_INTERFACE.PkgData pkgData = new Reflect_OBJ_INTERFACE.PkgData(
                Reflect_OBJ_INTERFACE.PkgName.CALL_SetProperty,
                Guid.NewGuid().ToString(),
                JToken.FromObject(rData)
                );

            string json = JsonConvert.SerializeObject(pkgData);
            Send(json);
        }

        #region FObj
        protected override void OnOpen()
        {
            Reflect_OBJ_INTERFACE.PkgData pkgData = new Reflect_OBJ_INTERFACE.PkgData(
                Reflect_OBJ_INTERFACE.PkgName.CALL_GetAllProperties,
                Guid.NewGuid().ToString(),
                null
                );
            string json = JsonConvert.SerializeObject(pkgData);
            Send(json);
        }
        protected override void OnMessage(string msg)
        {
            var pkgData = JsonConvert.DeserializeObject<Reflect_OBJ_INTERFACE.PkgData>(msg);

            var pkgName = Enum.Parse(typeof(Reflect_OBJ_INTERFACE.PkgName), pkgData.name);

            switch (pkgName)
            {
                case Reflect_OBJ_INTERFACE.PkgName.PUSH_PropertyChanged:
                case Reflect_OBJ_INTERFACE.PkgName.CALL_GetAllProperties:
                    {
                        ignoreSet = true;

                        var rData = pkgData.data.ToObject<Reflect_OBJ_INTERFACE.ReflectData>();

                        var node = COMMON.FindNode(rootNode, rData.name);
                        if (node == null)
                        {
                            //异常
                            //服务器乱发过来
                            break;
                        }

                        string json = rData.data.ToString();
                        JsonConvert.PopulateObject(json, node.Obj);


                        ignoreSet = false;
                    }
                    break;
                case Reflect_OBJ_INTERFACE.PkgName.PUSH_Event:
                    {
                        if (InterfaceType == null)
                            return;

                        var rData = pkgData.data.ToObject<Reflect_OBJ_INTERFACE.ReflectData>();

                        var anyEvent = anyEvents.Find(ae => ae.name == rData.name);
                        if (anyEvent == null)
                            return;//异常！！！


                        //触发事件！！！
                        var methodInfo = GetType().GetMethod(anyEvent.triggerName);
                        if (methodInfo == null)
                        {
                            throw new Exception($"客户端 {GetType()} 忘记写 {anyEvent.triggerName}");
                        }
                        var obj = rData.data.ToObject(anyEvent.retType);
                        //出错，就提示，肯定是客户端忘记写 "Trigger_XXXX"
                        methodInfo.Invoke(this, new object[] { obj });
                    }
                    break;

                case Reflect_OBJ_INTERFACE.PkgName.CALL_MethodInvoke:
                    {
                        if (InterfaceType == null)
                            return;

                        var reponse = pkgData.data.ToObject<Reflect_OBJ_INTERFACE.ReflectData>();
                        var anyCall = anyCalls.Find(ac => ac.name == reponse.name);
                        if (anyCall == null)
                        {
                            //异常！！！应该有[Call]注明返回的类型
                            throw new Exception($"{reponse.name} 异常！！！应该有[Call]注明返回的类型");
                        }

                        var retData = reponse.data.ToObject(anyCall.retType);
                        var anyMethodInvoke = anyMethodInvokes.Find(ami => ami.guid == pkgData.guid);

                        if (anyMethodInvoke == null)
                        {
                            //异常！！！应该有对应的回调
                            throw new Exception($"{reponse.name} 异常！！！应该有回调");
                        }

                        anyMethodInvoke.asyncDelegate.Invoke(anyMethodInvoke.asyncContext, retData);
                        anyMethodInvokes.Remove(anyMethodInvoke);
                        
                    }
                    break;
            }
        }


        protected void Call(string methodName, object parameters, AsyncCBHandler asyncDelegate, object asyncContext)
        {
            var rData = new Reflect_OBJ_INTERFACE.ReflectData()
            {
                name = methodName,
                data = parameters == null ? null : JObject.FromObject(parameters)
            };
            var pkgData = new Reflect_OBJ_INTERFACE.PkgData(Reflect_OBJ_INTERFACE.PkgName.CALL_MethodInvoke, Guid.NewGuid().ToString(), JObject.FromObject(rData));
            anyMethodInvokes.Add(new AnyMethodInvoke() { guid = pkgData.guid, asyncDelegate = asyncDelegate, asyncContext = asyncContext });

            string json = JsonConvert.SerializeObject(pkgData);
            Send(json);
        }
        protected void Call(string methodName, object parameters)
        {
            Call(methodName, parameters, null, null);
        }
        protected void Call(string methodName)
        {
            Call(methodName, null, null, null);
        }
        #endregion

        public void Dispose()
        {
            if (typeof(INotifyPropertyChanged).IsAssignableFrom(InterfaceType))
            {
                //继承了INotifyPropertyChanged
                ((INotifyPropertyChanged)(this)).PropertyChanged -= rootNode.PropertyChanged;
            }

            //释放subProperties
            COMMON.NodeDispose(rootNode);
        }
    }
}
